from inspect import isclass
from typing import Callable, Mapping, MutableMapping, cast, overload

from promplate.chain.callback import BaseCallback
from promplate.llm.base import Complete

from ..llm.base import *
from ..llm.base import AsyncGenerate, AsyncIterable, Generate, Iterable
from ..prompt.template import Context, Loader, SafeChainMapContext, Template
from .callback import BaseCallback, Callback
from .utils import iterate, resolve


class ChainContext(SafeChainMapContext):
    @overload
    def __init__(self):
        ...

    @overload
    def __init__(self, least: MutableMapping | None = None):
        ...

    @overload
    def __init__(self, least: MutableMapping | None = None, *maps: Mapping):
        ...

    def __init__(self, least: MutableMapping | None = None, *maps: Mapping):
        super().__init__({} if least is None else least, *maps)  # type: ignore

    @classmethod
    def ensure(cls, context):
        return context if isinstance(context, cls) else cls(context)

    @property
    def result(self):
        return self.__getitem__("__result__")

    @result.setter
    def result(self, result):
        self.__setitem__("__result__", result)

    @result.deleter
    def result(self):
        self.__delitem__("__result__")


Process = Callable[[ChainContext], Context | None]

AsyncProcess = Callable[[ChainContext], Awaitable[Context | None]]


class AbstractChain(Protocol):
    def invoke(
        self,
        context: Context | None = None,
        /,
        complete: Complete | None = None,
        **config,
    ) -> ChainContext:
        ...

    async def ainvoke(
        self,
        context: Context | None = None,
        /,
        complete: Complete | AsyncComplete | None = None,
        **config,
    ) -> ChainContext:
        ...

    def stream(
        self,
        context: Context | None = None,
        /,
        generate: Generate | None = None,
        **config,
    ) -> Iterable[ChainContext]:
        ...

    def astream(
        self,
        context: Context | None = None,
        /,
        generate: Generate | AsyncGenerate | None = None,
        **config,
    ) -> AsyncIterable[ChainContext]:
        ...


def ensure_callbacks(callbacks: list[BaseCallback | type[BaseCallback]]) -> list[BaseCallback]:
    return [i() if isclass(i) else i for i in callbacks]


class Interruptable(AbstractChain, Protocol):
    def _invoke(
        self,
        context: ChainContext,
        /,
        complete: Complete | None,
        callbacks: list[BaseCallback],
        **config,
    ):
        ...

    async def _ainvoke(
        self,
        context: ChainContext,
        /,
        complete: Complete | AsyncComplete | None,
        callbacks: list[BaseCallback],
        **config,
    ):
        ...

    def _stream(
        self,
        context: ChainContext,
        /,
        generate: Generate | None,
        callbacks: list[BaseCallback],
        **config,
    ) -> Iterable:
        ...

    def _astream(
        self,
        context: ChainContext,
        /,
        generate: Generate | AsyncGenerate | None,
        callbacks: list[BaseCallback],
        **config,
    ) -> AsyncIterable:
        ...

    callbacks: list[BaseCallback | type[BaseCallback]]

    def enter(self, context: Context | None, config: Context):
        callbacks: list[BaseCallback] = ensure_callbacks(self.callbacks)
        for callback in callbacks:
            context, config = callback.on_enter(self, context, config)
        return context, config, callbacks

    def leave(self, context: ChainContext, config: Context, callbacks: list[BaseCallback]):
        for callback in reversed(callbacks):
            context, config = callback.on_leave(self, context, config)
        return context, config

    def add_pre_processes(self, *processes: Process | AsyncProcess):
        self.callbacks.extend(Callback(pre_process=i) for i in processes)
        return self

    def add_mid_processes(self, *processes: Process | AsyncProcess):
        self.callbacks.extend(Callback(mid_process=i) for i in processes)
        return self

    def add_end_processes(self, *processes: Process | AsyncProcess):
        self.callbacks.extend(Callback(end_process=i) for i in processes)
        return self

    def add_callbacks(self, *callbacks: BaseCallback | type[BaseCallback]):
        self.callbacks.extend(callbacks)
        return self

    def pre_process(self, process: Process | AsyncProcess):
        self.add_pre_processes(process)
        return process

    def mid_process(self, process: Process | AsyncProcess):
        self.add_mid_processes(process)
        return process

    def end_process(self, process: Process | AsyncProcess):
        self.add_end_processes(process)
        return process

    def callback(self, callback: BaseCallback | type[BaseCallback]):
        self.add_callbacks(callback)
        return callback

    @staticmethod
    def _apply_pre_processes(context: ChainContext, callbacks: list[BaseCallback]):
        for callback in callbacks:
            context |= cast(Context, callback.pre_process(context) or {})

    @staticmethod
    def _apply_mid_processes(context: ChainContext, callbacks: list[BaseCallback]):
        for callback in callbacks:
            context |= cast(Context, callback.mid_process(context) or {})

    @staticmethod
    def _apply_end_processes(context: ChainContext, callbacks: list[BaseCallback]):
        for callback in callbacks:
            context |= cast(Context, callback.end_process(context) or {})

    @staticmethod
    async def _apply_async_pre_processes(context: ChainContext, callbacks: list[BaseCallback]):
        for callback in callbacks:
            context |= cast(Context, await resolve(callback.pre_process(context)) or {})

    @staticmethod
    async def _apply_async_mid_processes(context: ChainContext, callbacks: list[BaseCallback]):
        for callback in callbacks:
            context |= cast(Context, await resolve(callback.mid_process(context)) or {})

    @staticmethod
    async def _apply_async_end_processes(context: ChainContext, callbacks: list[BaseCallback]):
        for callback in callbacks:
            context |= cast(Context, await resolve(callback.end_process(context)) or {})

    def invoke(self, context=None, /, complete=None, **config) -> ChainContext:
        context, config, callbacks = self.enter(context, config)
        context = ChainContext.ensure(context)

        try:
            self._invoke(ChainContext(context, self.context), complete, callbacks, **config)
        except Jump as jump:
            context, config = self.leave(context, config, callbacks)
            if jump.out_of is None or jump.out_of is self:
                if jump.into is not None:
                    jump.into.invoke(context, complete, **config)
            else:
                raise jump from None
        else:
            context, config = self.leave(context, config, callbacks)

        return context

    async def ainvoke(self, context=None, /, complete=None, **config) -> ChainContext:
        context, config, callbacks = self.enter(context, config)
        context = ChainContext.ensure(context)

        try:
            await self._ainvoke(ChainContext(context, self.context), complete, callbacks, **config)
        except Jump as jump:
            context, config = self.leave(context, config, callbacks)
            if jump.out_of is None or jump.out_of is self:
                if jump.into is not None:
                    await jump.into.ainvoke(context, complete, **config)
            else:
                raise jump from None
        else:
            context, config = self.leave(context, config, callbacks)

        return context

    def stream(self, context=None, /, generate=None, **config) -> Iterable[ChainContext]:
        context, config, callbacks = self.enter(context, config)
        context = ChainContext.ensure(context)

        try:
            for _ in self._stream(ChainContext(context, self.context), generate, callbacks, **config):
                yield context
        except Jump as jump:
            context, config = self.leave(context, config, callbacks)
            if jump.out_of is None or jump.out_of is self:
                if jump.into is not None:
                    yield from jump.into.stream(context, generate, **config)
            else:
                raise jump from None
        else:
            context, config = self.leave(context, config, callbacks)

    async def astream(self, context=None, /, generate=None, **config) -> AsyncIterable[ChainContext]:
        context, config, callbacks = self.enter(context, config)
        context = ChainContext.ensure(context)

        try:
            async for _ in self._astream(ChainContext(context, self.context), generate, callbacks, **config):
                yield context
        except Jump as jump:
            context, config = self.leave(context, config, callbacks)
            if jump.out_of is None or jump.out_of is self:
                if jump.into is not None:
                    async for i in jump.into.astream(context, generate, **config):
                        yield i
            else:
                raise jump from None
        else:
            context, config = self.leave(context, config, callbacks)

    _context: Context | None

    @property
    def context(self):
        if self._context is None:
            self._context = {}
        return self._context

    @context.setter
    def context(self, context: Context | None):
        self._context = context

    @context.deleter
    def context(self):
        self._context = None


class Node(Loader, Interruptable):
    def __init__(
        self,
        template: Template | str,
        partial_context: Context | None = None,
        llm: LLM | None = None,
        **config,
    ):
        self.template = Template(template) if isinstance(template, str) else template
        self._context = partial_context
        self.callbacks: list[BaseCallback | type[BaseCallback]] = []
        self.llm = llm
        self.run_config = config

    def bind_llm(self, llm: LLM | None):
        self.llm = llm
        return llm

    def _invoke(self, context, /, complete, callbacks, **config):
        complete = self.llm.complete if self.llm else complete
        assert complete is not None

        prompt = self.render(context, callbacks)

        context.result = complete(prompt, **self.run_config, **config)

        self._apply_mid_processes(context, callbacks)

        self._apply_end_processes(context, callbacks)

    def _stream(self, context, /, generate, callbacks, **config):
        generate = self.llm.generate if self.llm else generate
        assert generate is not None

        prompt = self.render(context, callbacks)

        context.result = ""
        for delta in generate(prompt, **self.run_config, **config):  # type: ignore
            context.result += delta
            self._apply_mid_processes(context, callbacks)
            yield

        self._apply_end_processes(context, callbacks)

    async def _ainvoke(self, context, /, complete, callbacks, **config):
        complete = self.llm.complete if self.llm else complete
        assert complete is not None

        prompt = await self.arender(context, callbacks)

        context.result = await resolve(complete(prompt, **self.run_config, **config))

        await self._apply_async_mid_processes(context, callbacks)

        await self._apply_async_end_processes(context, callbacks)

    async def _astream(self, context, /, generate, callbacks, **config):
        generate = self.llm.generate if self.llm else generate
        assert generate is not None

        prompt = await self.arender(context, callbacks)

        context.result = ""
        async for delta in iterate(generate(prompt, **self.run_config, **config)):
            context.result += delta
            await self._apply_async_mid_processes(context, callbacks)
            yield

        await self._apply_async_end_processes(context, callbacks)

    @staticmethod
    def _get_chain_type():
        return Chain

    def __add__(self, chain: AbstractChain):
        if isinstance(chain, Chain):
            return self._get_chain_type()(self, *chain)
        else:
            return self._get_chain_type()(self, chain)

    def render(self, context: Context | None = None, callbacks: list[BaseCallback] | None = None):
        if callbacks is None:
            callbacks = ensure_callbacks(self.callbacks)
        context = ChainContext(context, self.context)
        self._apply_pre_processes(context, callbacks)
        return self.template.render(context)

    async def arender(self, context: Context | None = None, callbacks: list[BaseCallback] | None = None):
        if callbacks is None:
            callbacks = ensure_callbacks(self.callbacks)
        context = ChainContext(context, self.context)
        await self._apply_async_pre_processes(context, callbacks)
        return await self.template.arender(context)

    def __str__(self):
        return f"</{self.name}/>"


class Loop(Interruptable):
    def __init__(self, chain: AbstractChain, partial_context: Context | None = None):
        self.chain = chain
        self._context = partial_context
        self.callbacks: list[BaseCallback | type[BaseCallback]] = []

    def _invoke(self, context, /, complete, callbacks, **config):
        while True:
            self._apply_pre_processes(context, callbacks)
            self.chain.invoke(context, complete, **config)
            self._apply_mid_processes(context, callbacks)
            self._apply_end_processes(context, callbacks)

    async def _ainvoke(self, context, /, complete, callbacks, **config):
        while True:
            await self._apply_async_pre_processes(context, callbacks)
            await self.chain.ainvoke(context, complete, **config)
            await self._apply_async_mid_processes(context, callbacks)
            await self._apply_async_end_processes(context, callbacks)

    def _stream(self, context, /, generate, callbacks, **config):
        while True:
            self._apply_pre_processes(context, callbacks)
            for _ in self.chain.stream(context, generate, **config):
                self._apply_mid_processes(context, callbacks)
                yield
            self._apply_end_processes(context, callbacks)

    async def _astream(self, context, /, generate, callbacks, **config):
        while True:
            await self._apply_async_pre_processes(context, callbacks)
            async for _ in self.chain.astream(context, generate, **config):
                await self._apply_async_mid_processes(context, callbacks)
                yield
            await self._apply_async_end_processes(context, callbacks)


class Chain(Interruptable):
    def __init__(self, *nodes: AbstractChain, partial_context: Context | None = None):
        self.nodes = list(nodes)
        self._context = partial_context
        self.callbacks: list[BaseCallback | type[BaseCallback]] = []

    def __add__(self, chain: AbstractChain):
        if isinstance(chain, Node):
            return self.__class__(*self, chain)
        elif isinstance(chain, Chain):
            return self.__class__(*self, *chain)
        else:
            raise NotImplementedError

    def __iadd__(self, chain: AbstractChain):
        self.nodes.append(chain)
        return self

    def __iter__(self):
        return iter(self.nodes)

    def _invoke(self, context, /, complete, callbacks: list[BaseCallback], **config):
        self._apply_pre_processes(context, callbacks)
        for node in self.nodes:
            node.invoke(context, complete, **config)
            self._apply_mid_processes(context, callbacks)
        self._apply_end_processes(context, callbacks)

    async def _ainvoke(self, context, /, complete, callbacks: list[BaseCallback], **config):
        await self._apply_async_pre_processes(context, callbacks)
        for node in self.nodes:
            await node.ainvoke(context, complete, **config)
            await self._apply_async_mid_processes(context, callbacks)
        await self._apply_async_end_processes(context, callbacks)

    def _stream(self, context, /, generate, callbacks: list[BaseCallback], **config):
        self._apply_pre_processes(context, callbacks)
        for node in self.nodes:
            for _ in node.stream(context, generate, **config):
                self._apply_mid_processes(context, callbacks)
                yield
        self._apply_end_processes(context, callbacks)

    async def _astream(self, context, /, generate, callbacks: list[BaseCallback], **config):
        await self._apply_async_pre_processes(context, callbacks)
        for node in self.nodes:
            async for _ in node.astream(context, generate, **config):
                await self._apply_async_mid_processes(context, callbacks)
                yield
        await self._apply_async_end_processes(context, callbacks)

    def __repr__(self):
        return " + ".join(map(str, self.nodes))


class Jump(Exception):
    def __init__(self, into: Interruptable | None = None, out_of: Interruptable | None = None):
        self.into = into
        self.out_of = out_of

    def __str__(self):
        return f"{self.out_of} does not exist in the chain hierarchy"



from importlib.metadata import metadata
from typing import TYPE_CHECKING, Any

import openai
from openai import ChatCompletion, Completion  # type: ignore

from ...prompt.chat import Message, ensure
from ..base import *

meta = metadata("promplate")

if openai.api_key is None:
    openai.api_key = ""

openai.app_info = (openai.app_info or {}) | {  # type: ignore
    "name": "Promplate",
    "version": meta["version"],
    "url": meta["home-page"],
}


if TYPE_CHECKING:

    class Config(Configurable):
        def __init__(
            self,
            model: str,
            temperature: float | int | None = None,
            top_p: float | int | None = None,
            stop: str | list[str] | None = None,
            max_tokens: int | None = None,
            api_key: str | None = None,
            api_base: str | None = None,
            **other_config,
        ):
            self.model = model
            self.temperature = temperature
            self.top_p = top_p
            self.stop = stop
            self.max_tokens = max_tokens
            self.api_key = api_key
            self.api_base = api_base

            for key, val in other_config.items():
                setattr(self, key, val)

        def __setattr__(self, *_):
            ...

        def __getattr__(self, _):
            ...

else:
    Config = Configurable


class TextComplete(Config, Complete):
    def __call__(self, text: str, /, **config):
        config = self._config | config | {"stream": False, "prompt": text}
        result: Any = Completion.create(**config)
        return result["choices"][0]["text"]


class AsyncTextComplete(Config, AsyncComplete):
    async def __call__(self, text: str, /, **config):
        config = self._config | config | {"stream": False, "prompt": text}
        result: Any = await Completion.acreate(**config)
        return result["choices"][0]["text"]


class TextGenerate(Config, Generate):
    def __call__(self, text: str, /, **config):
        config = self._config | config | {"stream": True, "prompt": text}
        stream: Any = Completion.create(**config)
        for event in stream:
            yield event["choices"][0]["text"]


class AsyncTextGenerate(Config, AsyncGenerate):
    async def __call__(self, text: str, /, **config):
        config = self._config | config | {"stream": True, "prompt": text}
        stream: Any = await Completion.acreate(**config)
        async for event in stream:
            yield event["choices"][0]["text"]


class ChatComplete(Config, Complete):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": False, "messages": messages}
        result: Any = ChatCompletion.create(**config)
        return result["choices"][0]["message"]["content"]


class AsyncChatComplete(Config, AsyncComplete):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": False, "messages": messages}
        result: Any = await ChatCompletion.acreate(**config)
        return result["choices"][0]["message"]["content"]


class ChatGenerate(Config, Generate):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": True, "messages": messages}
        stream: Any = ChatCompletion.create(**config)
        for event in stream:
            delta: dict = event["choices"][0]["delta"]
            yield delta.get("content", "")


class AsyncChatGenerate(Config, AsyncGenerate):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._config | config | {"stream": True, "messages": messages}
        stream: Any = await ChatCompletion.acreate(**config)
        async for event in stream:
            delta: dict = event["choices"][0]["delta"]
            yield delta.get("content", "")



from functools import cached_property, wraps
from inspect import currentframe
from re import compile
from typing import Any, Callable, ParamSpec, TypeVar

split_template_tokens = compile(
    r"(\s{%-.*?-%}\s|\s{{-.*?-}}\s|\s{#-.*?-#}\s|\s{%-.*?%}|\s{{-.*?}}|\s{#-.*?#}|{%.*?-%}\s|{{.*?-}}\s|{#.*?-#}\s|{%.*?%}|{{.*?}}|{#.*?#})"
).split


var_name_checker = compile(r"[_a-zA-Z]\w*$")

is_message_start = compile(r"<\|\s?(user|system|assistant)\s?(\w{1,64})?\s?\|>")


def is_not_valid(name: str):
    return not var_name_checker.match(name)


def ensure_valid(name):
    if is_not_valid(name):
        raise NameError(name)


class AutoNaming:
    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj._bind_frame()
        return obj

    def _bind_frame(self) -> None:
        self._frame = currentframe()

    @cached_property
    def _name(self):
        f = self._frame
        if f and f.f_back and (frame := f.f_back.f_back):
            for name, var in frame.f_locals.items():
                if var is self:
                    return name

    @property
    def class_name(self):
        return self.__class__.__name__

    fallback_name = class_name

    @property
    def name(self):
        return self._name or self.fallback_name

    @name.setter
    def name(self, name):
        self._name = name
        self._frame = None

    @name.deleter
    def name(self):
        del self._name

    def __repr__(self):
        if self._name:
            return f"<{self.class_name} {self.name}>"
        else:
            return f"<{self.class_name}>"

    def __str__(self):
        return f"<{self.name}>"


P = ParamSpec("P")
T = TypeVar("T")


def cache_once(func: Callable[P, T]) -> Callable[P, T]:
    result = None

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        nonlocal result
        if result is None:
            result = func(*args, **kwargs)
        return result

    return wrapper


@cache_once
def get_builtins() -> dict[str, Any]:
    return __builtins__ if isinstance(__builtins__, dict) else __builtins__.__dict__



from copy import copy
from functools import cached_property
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Callable, ParamSpec, TypeVar

from openai import AsyncClient, Client  # type: ignore

from ...prompt.chat import Message, ensure
from ..base import *

P = ParamSpec("P")
T = TypeVar("T")


def same_params_as(_: Callable[P, Any]):
    def func(__init__: Callable[..., None]) -> Callable[P, None]:
        return __init__  # type: ignore

    return func


class Config(Configurable):
    def __init__(self, **config):
        super().__init__(**config)
        self._run_config = {}

    def bind(self, **run_config):
        obj = copy(self)
        obj._run_config = self._run_config | run_config
        return obj

    @property
    def _config(self):  # type: ignore
        return MappingProxyType(super()._config)

    @cached_property
    def client(self):
        return Client(**self._config)

    @cached_property
    def aclient(self):
        return AsyncClient(**self._config)


if TYPE_CHECKING:

    class ClientConfig(Config):
        @same_params_as(Client)
        def __init__(self, **config):
            ...

    class AsyncClientConfig(Config):
        @same_params_as(AsyncClient)
        def __init__(self, **config):
            ...

else:
    ClientConfig = AsyncClientConfig = Config


class TextComplete(ClientConfig):
    def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": False, "prompt": text}
        result = self.client.completions.create(**config)
        return result.choices[0].text


class AsyncTextComplete(AsyncClientConfig):
    async def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": False, "prompt": text}
        result = await self.aclient.completions.create(**config)
        return result.choices[0].text


class TextGenerate(ClientConfig):
    def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": True, "prompt": text}
        stream = self.client.completions.create(**config)
        for event in stream:
            yield event.choices[0].text


class AsyncTextGenerate(AsyncClientConfig):
    async def __call__(self, text: str, /, **config):
        config = self._run_config | config | {"stream": True, "prompt": text}
        stream = await self.aclient.completions.create(**config)
        async for event in stream:
            yield event.choices[0].text


class ChatComplete(ClientConfig):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": False, "messages": messages}
        result = self.client.chat.completions.create(**config)
        return result.choices[0].message.content


class AsyncChatComplete(AsyncClientConfig):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": False, "messages": messages}
        result = await self.aclient.chat.completions.create(**config)
        return result.choices[0].message.content


class ChatGenerate(ClientConfig):
    def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": True, "messages": messages}
        stream = self.client.chat.completions.create(**config)
        for event in stream:
            yield event.choices[0].delta.content or ""


class AsyncChatGenerate(AsyncClientConfig):
    async def __call__(self, messages: list[Message] | str, /, **config):
        messages = ensure(messages)
        config = self._run_config | config | {"stream": True, "messages": messages}
        stream = await self.aclient.chat.completions.create(**config)
        async for event in stream:
            yield event.choices[0].delta.content or ""


class SyncTextOpenAI(ClientConfig, LLM):
    complete: TextComplete = TextComplete.__call__  # type: ignore
    generate: TextGenerate = TextGenerate.__call__  # type: ignore


class AsyncTextOpenAI(AsyncClientConfig, LLM):
    complete: AsyncTextComplete = AsyncTextComplete.__call__  # type: ignore
    generate: AsyncTextGenerate = AsyncTextGenerate.__call__  # type: ignore


class SyncChatOpenAI(ClientConfig, LLM):
    complete: ChatComplete = ChatComplete.__call__  # type: ignore
    generate: ChatGenerate = ChatGenerate.__call__  # type: ignore


class AsyncChatOpenAI(AsyncClientConfig, LLM):
    complete: AsyncChatComplete = AsyncChatComplete.__call__  # type: ignore
    generate: AsyncChatGenerate = AsyncChatGenerate.__call__  # type: ignore



from typing import TYPE_CHECKING, Awaitable, Callable, Protocol

from ..prompt import Context

if TYPE_CHECKING:
    from .node import AsyncProcess, ChainContext, Interruptable, Process


class BaseCallback(Protocol):
    def pre_process(self, context: "ChainContext") -> Context | Awaitable[Context | None] | None:
        pass

    def mid_process(self, context: "ChainContext") -> Context | Awaitable[Context | None] | None:
        pass

    def end_process(self, context: "ChainContext") -> Context | Awaitable[Context | None] | None:
        pass

    def on_enter(self, node: "Interruptable", context: Context | None, config: Context) -> tuple[Context | None, Context]:
        return context, config

    def on_leave(self, node: "Interruptable", context: "ChainContext", config: Context) -> tuple["ChainContext", Context]:
        return context, config


class Callback(BaseCallback):
    def __init__(
        self,
        *,
        pre_process: "Process | AsyncProcess | None" = None,
        mid_process: "Process | AsyncProcess | None" = None,
        end_process: "Process | AsyncProcess | None" = None,
        on_enter: Callable[["Interruptable", Context | None, Context], tuple[Context | None, Context]] | None = None,
        on_leave: Callable[["Interruptable", "ChainContext", Context], tuple["ChainContext", Context]] | None = None,
    ) -> None:
        self._pre_process = pre_process
        self._mid_process = mid_process
        self._end_process = end_process
        self._on_enter = on_enter
        self._on_leave = on_leave

    def pre_process(self, context):
        if self._pre_process is not None:
            return self._pre_process(context)

    def mid_process(self, context):
        if self._mid_process is not None:
            return self._mid_process(context)

    def end_process(self, context):
        if self._end_process is not None:
            return self._end_process(context)

    def on_enter(self, node, context, config):
        if self._on_enter is not None:
            return self._on_enter(node, context, config)
        return context, config

    def on_leave(self, node, context, config):
        if self._on_leave is not None:
            return self._on_leave(node, context, config)
        return context, config
